from dataclasses import dataclass, field
from typing import List, Literal, Optional

from pydantic import BaseModel, field_validator


class Offsets(BaseModel):
    # in radians in FLU
    roll: Optional[float] = None
    pitch: Optional[float] = None
    yaw: Optional[float] = None
    forward: Optional[float] = None
    left: Optional[float] = None
    up: Optional[float] = None


class PTZLimits(BaseModel):
    min: float  # in radians if pan/tilt, fraction if zoom
    max: float  # in radians if pan/tilt, fraction if zoom


class PTZComponent(BaseModel):
    move_type: Literal["pan", "tilt", "zoom"]
    home: float  # in radians if pan/tilt, fraction if zoom
    joint_offsets: Optional[Offsets] = None
    joint_limits: Optional[PTZLimits] = None
    joystick_axis_index: Optional[int] = None
    reverse_joystick_input: bool = False


class PTZControlSettings(BaseModel):
    absolute_move_tolerance: Optional[float] = 0.0175  # radians
    query_timeout: Optional[int] = 2  # seconds
    check_camera_rate: Optional[float] = 1.0  # period in seconds
    publish_rate: Optional[float] = 10.0  # period in seconds
    wsdl_dir: Optional[str] = "/home/ros/.local/lib/python3.12/site-packages/wsdl"
    continuous_state_publishing: Optional[bool] = True
    # Whether to transform the camera from FRD (for ptz) to FLU
    frd_to_flu_transform: bool = True


class PTZOnvif(BaseModel):
    ip_address: str
    port: int
    username: str
    password: str
    # List of PTZ components that the camera has, order is important as it will determine joint hierarchy
    ptz_components: List[PTZComponent]
    settings: PTZControlSettings = PTZControlSettings()

    # Ensure that only one of each ptz move type is defined
    @field_validator("ptz_components")
    def check_unique_move_types(cls, v):
        move_types = [component.move_type for component in v]
        if len(move_types) != len(set(move_types)):
            raise ValueError("Duplicate move types in PTZ components")
        return v

    @field_validator("ptz_components")
    def check_has_components(cls, v):
        if len(v) == 0:
            raise ValueError("No PTZ components defined")
        return v


class Camera(BaseModel):
    # This will become the name of frame-id, ros topic and webrtc stream
    name: str
    # Used to order the stream in the UI
    order: int
    # An array of gstream source / transform elements e.g.
    # ["v4l2src", "video/x-raw, format=RGB,width=1920,height=1080"]
    elements: List[str]

    pixel_width: int
    pixel_height: int
    # Published ros topic names
    camera_frame_topic: Optional[str]
    camera_info_topic: Optional[str]
    # Camera intrinsics and distortion parameters from callibration
    k_intrinsic: Optional[List[float]] = field(
        default_factory=lambda: [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]
    )
    distortion_parameters: Optional[List[float]] = field(
        default_factory=lambda: [0.0, 0.0, 0.0, 0.0, 0.0]
    )
    distortion_model: str = "plumb_bob"
    # If false, it expects the camera_info topic published somewhere else e.g. from rosbag
    publish_camera_info: bool = True
    # Free scaling parameter for undistorted image between 0 (all pixels are valid), and 1 (all source pixels are retained i.e. max distortion FOV)
    distortion_kmat_alpha: float = 0.5
    # The camera's position relative to the vessel base_link
    offsets: Optional[Offsets] = None
    type: str = "color"
    ros_throttle_time: float = 0.0000001
    # Whether to undistort the image in gstreamer pipeline before publishing
    undistort_image: bool = False
    # Whether the camera has a pan-tilt-zoom system enabled.
    ptz: Optional[PTZOnvif] = None


class CameraOverride(BaseModel):
    # This will become the name of frame-id, ros topic and webrtc stream
    name: Optional[str]
    order: Optional[int]
    # An array of gstream source / transform elements e.g.
    # ["v4l2src", "video/x-raw, format=RGB,width=1920,height=1080"]
    elements: Optional[List[str]]
    pixel_width: Optional[int]
    pixel_height: Optional[int]
    # Published ros topic names
    camera_frame_topic: Optional[str]
    camera_info_topic: Optional[str]
    # Camera intrinsics and distortion parameters from callibration
    k_intrinsic: Optional[List[float]] = None
    distortion_parameters: Optional[List[float]] = None
    distortion_model: str = "plumb_bob"
    # The camera's position relative to the vessel base_link
    offsets: Optional[Offsets] = None
    type: Optional[str] = "color"
    ros_throttle_time: Optional[float] = 0.0000001
    # Whether to undistort the image in gstreamer pipeline before publishing
    undistort_image: Optional[bool] = False


class GreenstreamConfig(BaseModel):
    cameras: List[Camera]
    camera_overrides: Optional[List[Optional[CameraOverride]]] = None
    signalling_server_port: int = 8443
    namespace_vessel: str = "vessel_1"
    namespace_application: str = "greenstream"
    ui_port: int = 8000
    mode: str = "simulator"
    debug: bool = False
    diagnostics_topic: str = "diagnostics"
    persistent_params_config_path: str = "/home/ros/.config/greenstream.yml"



from typing import List, Optional

from greenstream_config.types import Camera, CameraOverride


def merge_cameras(
    cameras: List[Camera], overrides: List[Optional[CameraOverride]]
) -> List[Camera]:
    """
    The gama_vessel may contain camera overrides.
    If it does, we need to merge them with the default cameras.
    """

    if overrides is None:
        return cameras

    overriden_cameras: List[Camera] = cameras
    for idx, override in enumerate(overrides):
        # If override is null/None, we don't want to override the default camera
        if override:
            if idx < len(overriden_cameras):

                # applies changes to the camera only if attribute is not None
                overriden_cameras[idx] = Camera(
                    name=override.name or cameras[idx].name,
                    order=override.order or cameras[idx].order,
                    elements=override.elements or cameras[idx].elements,
                    offsets=override.offsets or cameras[idx].offsets,
                    pixel_width=override.pixel_width or cameras[idx].pixel_width,
                    pixel_height=override.pixel_height or cameras[idx].pixel_height,
                    k_intrinsic=override.k_intrinsic or cameras[idx].k_intrinsic,
                    distortion_model=override.distortion_model or cameras[idx].distortion_model,
                    distortion_parameters=override.distortion_parameters
                    or cameras[idx].distortion_parameters,
                    camera_frame_topic=override.camera_frame_topic
                    or cameras[idx].camera_frame_topic,
                    camera_info_topic=override.camera_info_topic or cameras[idx].camera_info_topic,
                    type=override.type or cameras[idx].type,
                    ros_throttle_time=override.ros_throttle_time or cameras[idx].ros_throttle_time,
                    undistort_image=override.undistort_image or cameras[idx].undistort_image,
                )

    return overriden_cameras



from greenstream_config.merge_cameras import merge_cameras
from greenstream_config.types import (
    Camera,
    CameraOverride,
    GreenstreamConfig,
    Offsets,
    PTZComponent,
    PTZControlSettings,
    PTZLimits,
    PTZOnvif,
)
from greenstream_config.urdf import get_camera_urdf, get_cameras_urdf

__all__ = [
    "GreenstreamConfig",
    "Camera",
    "CameraOverride",
    "Offsets",
    "PTZComponent",
    "PTZLimits",
    "PTZControlSettings",
    "PTZOnvif",
    "get_camera_urdf",
    "get_cameras_urdf",
    "merge_cameras",
]
